@@ -1,7 +1,7 @@
 /* ta.h */
 
 /*
- * Copyright (C) 2011 by Werner Lemberg.
+ * Copyright (C) 2011-2012 by Werner Lemberg.
  *
  * This file is part of the ttfautohint library, and may only be used,
  * modified, and distributed under the terms given in `COPYING'.  By
@@ -103,12 +103,17 @@ typedef struct GLYPH_
   FT_ULong ins_len; /* number of new instructions */
   FT_Byte* ins_buf; /* new instruction data */
 
+  FT_Short num_contours; /* >= 0 for simple glyphs */
+  FT_UShort num_points; /* number of points in a simple glyph */
+
   FT_UShort num_components;
   FT_UShort* components; /* the subglyph indices of a composite glyph */
 
-  FT_UShort num_endpoints;
-  FT_UShort* endpoints; /* the endpoints of all composite elements */
+  FT_UShort num_pointsums;
+  FT_UShort* pointsums; /* the pointsums of all composite elements */
                         /* (after walking recursively over all subglyphs) */
+
+  FT_UShort num_composite_contours; /* after recursion */
 } GLYPH;
 
 /* a representation of the data in the `glyf' table */
