@@ -1,7 +1,7 @@
 /* taglyf.c */
 
 /*
- * Copyright (C) 2011 by Werner Lemberg.
+ * Copyright (C) 2011-2012 by Werner Lemberg.
  *
  * This file is part of the ttfautohint library, and may only be used,
  * modified, and distributed under the terms given in `COPYING'.  By
@@ -42,20 +42,16 @@ TA_sfnt_build_glyf_hints(SFNT* sfnt,
 
 
 static FT_Error
-TA_glyph_parse_composite(GLYPH* glyph,
-                         FT_Byte* buf,
-                         FT_UShort num_glyphs,
-                         FT_ULong len)
+TA_glyph_get_components(GLYPH* glyph,
+                        FT_Byte* buf,
+                        FT_ULong len)
 {
-  FT_ULong flags_offset; /* after the loop, this is the offset */
-                         /* to the last element in the flags array */
   FT_UShort flags;
   FT_UShort component;
   FT_UShort* components_new;
 
   FT_Byte* p;
   FT_Byte* endp;
-  FT_ULong new_len;
 
 
   p = buf;
@@ -70,8 +66,6 @@ TA_glyph_parse_composite(GLYPH* glyph,
     if (p + 4 > endp)
       return FT_Err_Invalid_Table;
 
-    flags_offset = p - buf;
-
     flags = *(p++) << 8;
     flags += *(p++);
 
@@ -99,8 +93,6 @@ TA_glyph_parse_composite(GLYPH* glyph,
     else
       p += 2;
 
-    /* XXX adjust point indices for !ARGS_ARE_XY_VALUES */
-
     if (flags & WE_HAVE_A_SCALE)
       p += 2;
     else if (flags & WE_HAVE_AN_XY_SCALE)
@@ -109,36 +101,204 @@ TA_glyph_parse_composite(GLYPH* glyph,
       p += 8;
   } while (flags & MORE_COMPONENTS);
 
+  return TA_Err_Ok;
+}
+
+
+static FT_Error
+TA_glyph_parse_composite(GLYPH* glyph,
+                         FT_Byte* buf,
+                         FT_ULong len,
+                         FT_UShort num_glyphs)
+{
+  FT_ULong flags_offset; /* after the loop, this is the offset */
+                         /* to the last element in the flags array */
+  FT_UShort flags;
+
+  FT_Byte* p;
+  FT_Byte* q;
+
+
+  /* we allocate too large a buffer */
+  /* (including space for the new component */
+  /* and possible argument size changes for shifted point indices) */
+  /* and reallocate it later to its real size */
+  glyph->buf = (FT_Byte*)malloc(len + 8 + glyph->num_components * 2);
+  if (!glyph->buf)
+    return FT_Err_Out_Of_Memory;
+
+  p = buf;
+  q = glyph->buf;
+
+  /* copy header */
+  memcpy(q, p, 10);
+  p += 10;
+  q += 10;
+
+  /* if the composite glyph contains one or more contours, */
   /* we prepend a composite glyph component to call some bytecode */
   /* which eventually becomes the last glyph in the `glyf' table; */
   /* for convenience, however, it is not added to the `components' array */
   /* (doing so simplifies the conversion of point indices later on) */
+  if (glyph->num_composite_contours)
+  {
+    FT_Short x_min;
+    FT_Short x_max;
+    FT_Short y_min;
+    FT_Short y_max;
+    FT_Short x_offset;
+    FT_Short y_offset;
 
-  /* adjust glyph record length (6 bytes for the additional component) */
-  new_len = p - buf + 6;
 
-  glyph->flags_offset = flags_offset + 6;
+    /* the composite glyph's bounding box */
+    x_min = (FT_Short)((buf[2] << 8) + buf[3]);
+    y_min = (FT_Short)((buf[4] << 8) + buf[5]);
+    x_max = (FT_Short)((buf[6] << 8) + buf[7]);
+    y_max = (FT_Short)((buf[8] << 8) + buf[9]);
 
-  glyph->len1 = new_len;
+    /* use ARGS_ARE_WORDS only if necessary; */
+    /* note that the offset value of the component doesn't matter */
+    /* as long as it stays within the bounding box */
+    if (x_min <= 0 && x_max >= 0)
+      x_offset = 0;
+    else if (x_max < 0)
+      x_offset = x_max;
+    else
+      x_offset = x_min;
+
+    if (y_min <= 0 && y_max >= 0)
+      y_offset = 0;
+    else if (y_max < 0)
+      y_offset = y_max;
+    else
+      y_offset = y_min;
+
+    if (x_offset >= -128 && x_offset <= 127
+        && y_offset >= -128 && y_offset <= 127)
+    {
+      *(q++) = 0x00;
+      *(q++) = ARGS_ARE_XY_VALUES | MORE_COMPONENTS;
+      *(q++) = HIGH(num_glyphs - 1);
+      *(q++) = LOW(num_glyphs - 1);
+      *(q++) = x_offset;
+      *(q++) = y_offset;
+    }
+    else
+    {
+      *(q++) = 0x00;
+      *(q++) = ARGS_ARE_WORDS | ARGS_ARE_XY_VALUES | MORE_COMPONENTS;
+      *(q++) = HIGH(num_glyphs - 1);
+      *(q++) = LOW(num_glyphs - 1);
+      *(q++) = HIGH(x_offset);
+      *(q++) = LOW(x_offset);
+      *(q++) = HIGH(y_offset);
+      *(q++) = LOW(y_offset);
+    }
+  }
+
+  /* walk over component records */
+  do
+  {
+    flags_offset = q - glyph->buf;
+
+    *(q++) = *p;
+    flags = *(p++) << 8;
+    *(q++) = *p;
+    flags += *(p++);
+
+    /* copy component */
+    *(q++) = *(p++);
+    *(q++) = *(p++);
+
+    if (flags & ARGS_ARE_XY_VALUES)
+    {
+      /* copy offsets */
+      *(q++) = *(p++);
+      *(q++) = *(p++);
+
+      if (flags & ARGS_ARE_WORDS)
+      {
+        *(q++) = *(p++);
+        *(q++) = *(p++);
+      }
+    }
+    else
+    {
+      /* handle point numbers */
+      FT_UShort arg1;
+      FT_UShort arg2;
+      FT_UShort i;
+
+
+      if (flags & ARGS_ARE_WORDS)
+      {
+        arg1 = *(p++) >> 8;
+        arg1 += *(p++);
+        arg2 = *(p++) >> 8;
+        arg2 += *(p++);
+      }
+      else
+      {
+        arg1 = *(p++);
+        arg2 = *(p++);
+      }
+
+      /* adjust point numbers */
+      /* (see `TA_adjust_point_index' in `tabytecode.c' for more) */
+      for (i = 0; i < glyph->num_pointsums; i++)
+        if (arg1 < glyph->pointsums[i])
+          break;
+      arg1 += i;
+
+      for (i = 0; i < glyph->num_pointsums; i++)
+        if (arg2 < glyph->pointsums[i])
+          break;
+      arg2 += i;
+
+      if (arg1 <= 0xFF && arg2 <= 0xFF)
+      {
+        glyph->buf[flags_offset + 1] &= ~ARGS_ARE_WORDS;
+
+        *(q++) = arg1;
+        *(q++) = arg2;
+      }
+      else
+      {
+        glyph->buf[flags_offset + 1] |= ARGS_ARE_WORDS;
+
+        *(q++) = HIGH(arg1);
+        *(q++) = LOW(arg1);
+        *(q++) = HIGH(arg2);
+        *(q++) = LOW(arg2);
+      }
+    }
+
+    /* copy scaling arguments */
+    if (flags & (WE_HAVE_A_SCALE | WE_HAVE_AN_XY_SCALE | WE_HAVE_A_2X2))
+    {
+      *(q++) = *(p++);
+      *(q++) = *(p++);
+    }
+    if (flags & (WE_HAVE_AN_XY_SCALE | WE_HAVE_A_2X2))
+    {
+      *(q++) = *(p++);
+      *(q++) = *(p++);
+    }
+    if (flags & WE_HAVE_A_2X2)
+    {
+      *(q++) = *(p++);
+      *(q++) = *(p++);
+      *(q++) = *(p++);
+      *(q++) = *(p++);
+    }
+  } while (flags & MORE_COMPONENTS);
+
+  glyph->len1 = q - glyph->buf;
   /* glyph->len2 = 0; */
-  glyph->buf = (FT_Byte*)malloc(new_len);
-  if (!glyph->buf)
-    return FT_Err_Out_Of_Memory;
-
-  /* copy record without instructions (if any) */
-  /* and construct additional component */
-
-  memcpy(glyph->buf, buf, 10); /* header */
-
-  glyph->buf[10] = 0x00; /* additional component */
-  glyph->buf[11] = ARGS_ARE_XY_VALUES | MORE_COMPONENTS;
-  glyph->buf[12] = HIGH(num_glyphs - 1);
-  glyph->buf[13] = LOW(num_glyphs - 1);
-  glyph->buf[14] = 0x00;
-  glyph->buf[15] = 0x00;
-
-  memcpy(glyph->buf + 16, buf + 10, new_len - 6 - 10); /* the rest */
+  glyph->flags_offset = flags_offset;
+  glyph->buf = (FT_Byte*)realloc(glyph->buf, glyph->len1);
 
+  /* we discard instructions (if any) */
   glyph->buf[glyph->flags_offset] &= ~(WE_HAVE_INSTR >> 8);
 
   return TA_Err_Ok;
@@ -148,14 +308,12 @@ TA_glyph_parse_composite(GLYPH* glyph,
 static FT_Error
 TA_glyph_parse_simple(GLYPH* glyph,
                       FT_Byte* buf,
-                      FT_UShort num_contours,
                       FT_ULong len)
 {
   FT_ULong ins_offset;
   FT_Byte* flags_start;
 
   FT_UShort num_ins;
-  FT_UShort num_pts;
 
   FT_ULong flags_size; /* size of the flags array */
   FT_ULong xy_size; /* size of x and y coordinate arrays together */
@@ -169,7 +327,7 @@ TA_glyph_parse_simple(GLYPH* glyph,
   p = buf;
   endp = buf + len;
 
-  ins_offset = 10 + num_contours * 2;
+  ins_offset = 10 + glyph->num_contours * 2;
 
   p += ins_offset;
 
@@ -185,16 +343,11 @@ TA_glyph_parse_simple(GLYPH* glyph,
   if (p > endp)
     return FT_Err_Invalid_Table;
 
-  /* get number of points from last outline point */
-  num_pts = buf[ins_offset - 2] << 8;
-  num_pts += buf[ins_offset - 1];
-  num_pts++;
-
   flags_start = p;
   xy_size = 0;
   i = 0;
 
-  while (i < num_pts)
+  while (i < glyph->num_points)
   {
     FT_Byte flags;
     FT_Byte x_short;
@@ -224,7 +377,7 @@ TA_glyph_parse_simple(GLYPH* glyph,
 
       count += *(p++);
 
-      if (i + count > num_pts)
+      if (i + count > glyph->num_points)
         return FT_Err_Invalid_Table;
     }
 
@@ -259,30 +412,33 @@ static FT_Error
 TA_iterate_composite_glyph(glyf_Data* data,
                            FT_UShort* components,
                            FT_UShort num_components,
-                           FT_UShort** endpoints,
-                           FT_UShort* num_endpoints,
+                           FT_UShort** pointsums,
+                           FT_UShort* num_pointsums,
                            FT_UShort* num_composite_contours,
                            FT_UShort* num_composite_points)
 {
-  FT_UShort* endpoints_new;
+  FT_UShort* pointsums_new;
   FT_UShort i;
 
 
   /* save current state */
 
-  (*num_endpoints)++;
-  endpoints_new = (FT_UShort*)realloc(*endpoints,
-                                      *num_endpoints
+  if (*num_pointsums == 0xFFFF)
+    return FT_Err_Invalid_Table;
+
+  (*num_pointsums)++;
+  pointsums_new = (FT_UShort*)realloc(*pointsums,
+                                      *num_pointsums
                                       * sizeof (FT_UShort));
-  if (!endpoints_new)
+  if (!pointsums_new)
   {
-    (*num_endpoints)--;
+    (*num_pointsums)--;
     return FT_Err_Out_Of_Memory;
   }
   else
-    *endpoints = endpoints_new;
+    *pointsums = pointsums_new;
 
-  (*endpoints)[*num_endpoints - 1] = *num_composite_points;
+  (*pointsums)[*num_pointsums - 1] = *num_composite_points;
 
   for (i = 0; i < num_components; i++)
   {
@@ -301,8 +457,8 @@ TA_iterate_composite_glyph(glyf_Data* data,
       error = TA_iterate_composite_glyph(data,
                                          glyph->components,
                                          glyph->num_components,
-                                         endpoints,
-                                         num_endpoints,
+                                         pointsums,
+                                         num_pointsums,
                                          num_composite_contours,
                                          num_composite_points);
       if (error)
@@ -310,17 +466,13 @@ TA_iterate_composite_glyph(glyf_Data* data,
     }
     else
     {
-      FT_UShort num_contours;
-      FT_UShort endpoint;
+      /* no need for checking overflow of the number of contours */
+      /* since the number of points is always larger or equal */
+      if (*num_composite_points > 0xFFFF - glyph->num_points)
+        return FT_Err_Invalid_Table;
 
-
-      num_contours = glyph->buf[0] << 8;
-      num_contours += glyph->buf[1];
-      endpoint = glyph->buf[10 + (num_contours - 1) * 2] << 8;
-      endpoint += glyph->buf[10 + (num_contours - 1) * 2 + 1];
-
-      *num_composite_contours += num_contours;
-      *num_composite_points += endpoint + 1;
+      *num_composite_contours += glyph->num_contours;
+      *num_composite_points += glyph->num_points;
     }
   }
 
@@ -328,10 +480,8 @@ TA_iterate_composite_glyph(glyf_Data* data,
 }
 
 
-/* this function actually stores `endpoint + 1' values */
-
 static FT_Error
-TA_sfnt_compute_composite_endpoints(SFNT* sfnt,
+TA_sfnt_compute_composite_pointsums(SFNT* sfnt,
                                     FONT* font)
 {
   SFNT_Table* glyf_table = &font->tables[sfnt->glyf_idx];
@@ -355,24 +505,26 @@ TA_sfnt_compute_composite_endpoints(SFNT* sfnt,
       error = TA_iterate_composite_glyph(data,
                                          glyph->components,
                                          glyph->num_components,
-                                         &glyph->endpoints,
-                                         &glyph->num_endpoints,
+                                         &glyph->pointsums,
+                                         &glyph->num_pointsums,
                                          &num_composite_contours,
                                          &num_composite_points);
       if (error)
         return error;
 
+      glyph->num_composite_contours = num_composite_contours;
+
       /* update maximum values, */
       /* including the subglyphs not in `components' array */
       /* (each of them has a single point in a single contour) */
-      if (num_composite_points + glyph->num_endpoints
+      if (num_composite_points + glyph->num_pointsums
           > sfnt->max_composite_points)
         sfnt->max_composite_points = num_composite_points
-                                     + glyph->num_endpoints;
-      if (num_composite_contours + glyph->num_endpoints
+                                     + glyph->num_pointsums;
+      if (num_composite_contours + glyph->num_pointsums
           > sfnt->max_composite_contours)
         sfnt->max_composite_contours = num_composite_contours
-                                       + glyph->num_endpoints;
+                                       + glyph->num_pointsums;
     }
   }
 
@@ -427,6 +579,8 @@ TA_sfnt_split_glyf_table(SFNT* sfnt,
   if (!data->glyphs)
     return FT_Err_Out_Of_Memory;
 
+  /* first loop over `loca' and `glyf' data */
+
   p = loca_table->buf;
 
   if (loca_format)
@@ -443,7 +597,6 @@ TA_sfnt_split_glyf_table(SFNT* sfnt,
     offset_next <<= 1;
   }
 
-  /* loop over `loca' and `glyf' data */
   for (i = 0; i < loop_count; i++)
   {
     GLYPH* glyph = &data->glyphs[i];
@@ -476,15 +629,92 @@ TA_sfnt_split_glyf_table(SFNT* sfnt,
     else
     {
       FT_Byte* buf;
-      FT_Short num_contours;
 
 
       /* check header size */
       if (len < 10)
         return FT_Err_Invalid_Table;
 
+      /* we need the number of contours and points for */
+      /* `TA_sfnt_compute_composite_pointsums' */
+      buf = glyf_table->buf + offset;
+      glyph->num_contours = (FT_Short)((buf[0] << 8) + buf[1]);
+
+      if (glyph->num_contours < 0)
+      {
+        error = TA_glyph_get_components(glyph, buf, len);
+        if (error)
+          return error;
+      }
+      else
+      {
+        FT_ULong off;
+
+
+        /* use the last contour's end point to compute number of points */
+        off = 10 + (glyph->num_contours - 1) * 2;
+        glyph->num_points = buf[off] << 8;
+        glyph->num_points += buf[off + 1] + 1;
+      }
+    }
+  }
+
+  if (sfnt->max_components)
+  {
+    error = TA_sfnt_compute_composite_pointsums(sfnt, font);
+    if (error)
+      return error;
+  }
+
+  /* second loop over `loca' and `glyf' data */
+
+  p = loca_table->buf;
+
+  if (loca_format)
+  {
+    offset_next = *(p++) << 24;
+    offset_next += *(p++) << 16;
+    offset_next += *(p++) << 8;
+    offset_next += *(p++);
+  }
+  else
+  {
+    offset_next = *(p++) << 8;
+    offset_next += *(p++);
+    offset_next <<= 1;
+  }
+
+  for (i = 0; i < loop_count; i++)
+  {
+    GLYPH* glyph = &data->glyphs[i];
+    FT_ULong len;
+
+
+    offset = offset_next;
+
+    if (loca_format)
+    {
+      offset_next = *(p++) << 24;
+      offset_next += *(p++) << 16;
+      offset_next += *(p++) << 8;
+      offset_next += *(p++);
+    }
+    else
+    {
+      offset_next = *(p++) << 8;
+      offset_next += *(p++);
+      offset_next <<= 1;
+    }
+
+    len = offset_next - offset;
+    if (!len)
+      continue; /* empty glyph */
+    else
+    {
+      FT_Byte* buf;
+
+
       buf = glyf_table->buf + offset;
-      num_contours = (FT_Short)((buf[0] << 8) + buf[1]);
 
       /* We must parse the rest of the glyph record to get the exact */
       /* record length.  Since the `loca' table rounds record lengths */
@@ -493,10 +723,10 @@ TA_sfnt_split_glyf_table(SFNT* sfnt,
       /* possible otherwise to have more than 4 bytes of padding which */
       /* is more or less invalid. */
 
-      if (num_contours < 0)
-        error = TA_glyph_parse_composite(glyph, buf, data->num_glyphs, len);
+      if (glyph->num_contours < 0)
+        error = TA_glyph_parse_composite(glyph, buf, len, data->num_glyphs);
       else
-        error = TA_glyph_parse_simple(glyph, buf, num_contours, len);
+        error = TA_glyph_parse_simple(glyph, buf, len);
       if (error)
         return error;
     }
@@ -554,10 +784,6 @@ TA_sfnt_split_glyf_table(SFNT* sfnt,
     memcpy(glyph->ins_buf, bytecode, glyph->ins_len);
 
     sfnt->max_components += 1;
-
-    error = TA_sfnt_compute_composite_endpoints(sfnt, font);
-    if (error)
-      return error;
   }
 
   return TA_Err_Ok;
